@@ -82,6 +82,7 @@ fatal_1350(cPersistentObject *self, const char *caller, const char *detail)
 #endif
 
 static void ghostify(cPersistentObject*);
+static PyObject * pickle_slotnames(PyTypeObject *cls);
 
 static PyObject * convert_name(PyObject *name);
 
@@ -100,7 +101,7 @@ unghostify(cPersistentObject *self)
         {
             /* Create a node in the ring for this unghostified object. */
             self->cache->non_ghost_count++;
-            self->cache->total_estimated_size += 
+            self->cache->total_estimated_size +=
                 _estimated_size_in_bytes(self->estimated_size);
             ring_add(&self->cache->ring_home, &self->ring);
             Py_INCREF(self);
@@ -150,14 +151,15 @@ accessed(cPersistentObject *self)
 static void
 ghostify(cPersistentObject *self)
 {
-    PyObject **dictptr;
+    PyObject **dictptr, *slotnames;
+    PyObject *errtype, *errvalue, *errtb;
 
     /* are we already a ghost? */
     if (self->state == cPersistent_GHOST_STATE)
         return;
 
     /* Is it ever possible to not have a cache? */
-    if (self->cache == NULL) 
+    if (self->cache == NULL)
     {
         self->state = cPersistent_GHOST_STATE;
         return;
@@ -176,10 +178,12 @@ ghostify(cPersistentObject *self)
     /* If we're ghostifying an object, we better have some non-ghosts. */
     assert(self->cache->non_ghost_count > 0);
     self->cache->non_ghost_count--;
-    self->cache->total_estimated_size -= 
+    self->cache->total_estimated_size -=
         _estimated_size_in_bytes(self->estimated_size);
     ring_del(&self->ring);
     self->state = cPersistent_GHOST_STATE;
+
+    /* clear __dict__ */
     dictptr = _PyObject_GetDictPtr((PyObject *)self);
     if (dictptr && *dictptr)
     {
@@ -187,6 +191,60 @@ ghostify(cPersistentObject *self)
         *dictptr = NULL;
     }
 
+    /* clear all slots besides _p_*
+     * ( for backward-compatibility reason we do this only if class does not
+     *   override __new__ ) */
+    if (Py_TYPE(self)->tp_new == Pertype.tp_new)
+    {
+        /* later we might clear an AttributeError but
+         * if we have a pending exception that still needs to be
+         * raised so that we don't generate a SystemError.
+         */
+        PyErr_Fetch(&errtype, &errvalue, &errtb);
+
+        slotnames = pickle_slotnames(Py_TYPE(self));
+        if (slotnames && slotnames != Py_None)
+        {
+            int i;
+
+            for (i = 0; i < PyList_GET_SIZE(slotnames); i++)
+            {
+                PyObject *name;
+                char *cname;
+                int is_special;
+
+                name = PyList_GET_ITEM(slotnames, i);
+#ifdef PY3K
+                if (PyUnicode_Check(name))
+                {
+                    PyObject *converted = convert_name(name);
+                    cname = PyBytes_AS_STRING(converted);
+#else
+                if (PyBytes_Check(name))
+                {
+                    cname = PyBytes_AS_STRING(name);
+#endif
+                    is_special = !strncmp(cname, "_p_", 3);
+#ifdef PY3K
+                    Py_DECREF(converted);
+#endif
+                    if (is_special) /* skip persistent */
+                    {
+                        continue;
+                    }
+                }
+
+                /* NOTE: this skips our delattr hook */
+                if (PyObject_GenericSetAttr((PyObject *)self, name, NULL) < 0)
+                    /* delattr of non-set slot will raise AttributeError - we
+                     * simply ignore. */
+                    PyErr_Clear();
+            }
+        }
+        Py_XDECREF(slotnames);
+        PyErr_Restore(errtype, errvalue, errtb);
+    }
+
     /* We remove the reference to the just ghosted object that the ring
     * holds.  Note that the dictionary of oids->objects has an uncounted
     * reference, so if the ring's reference was the only one, this frees
@@ -212,7 +270,7 @@ changed(cPersistentObject *self)
         if (meth == NULL)
             return -1;
         arg = PyTuple_New(1);
-        if (arg == NULL) 
+        if (arg == NULL)
         {
             Py_DECREF(meth);
             return -1;
@@ -270,6 +328,8 @@ Per__p_deactivate(cPersistentObject *self)
             called directly. Methods that override this need to
             do the same! */
         ghostify(self);
+        if (PyErr_Occurred())
+            return NULL;
     }
 
     Py_INCREF(Py_None);
@@ -298,6 +358,8 @@ Per__p_invalidate(cPersistentObject *self)
         if (Per_set_changed(self, NULL) < 0)
             return NULL;
         ghostify(self);
+        if (PyErr_Occurred())
+            return NULL;
     }
     Py_INCREF(Py_None);
     return Py_None;
@@ -317,7 +379,7 @@ pickle_slotnames(PyTypeObject *cls)
             return NULL;
         if (n)
             slotnames = Py_None;
-        
+
         Py_INCREF(slotnames);
         return slotnames;
     }
@@ -534,11 +596,13 @@ pickle___setstate__(PyObject *self, PyObject *state)
     if (state != Py_None)
     {
         PyObject **dict;
+        PyObject *items;
         PyObject *d_key, *d_value;
         Py_ssize_t i;
+        int len;
 
         dict = _PyObject_GetDictPtr(self);
-        
+
         if (!dict)
         {
             PyErr_SetString(PyExc_TypeError,
@@ -555,17 +619,70 @@ pickle___setstate__(PyObject *self, PyObject *state)
 
         PyDict_Clear(*dict);
 
-        i = 0;
-        while (PyDict_Next(state, &i, &d_key, &d_value)) {
-            /* normally the keys for instance attributes are
-               interned.  we should try to do that here. */
-            if (NATIVE_CHECK_EXACT(d_key)) {
-                Py_INCREF(d_key);
-                INTERN_INPLACE(&d_key);
-                Py_DECREF(d_key);
+        if (PyDict_CheckExact(state))
+        {
+            i = 0;
+            while (PyDict_Next(state, &i, &d_key, &d_value)) {
+                /* normally the keys for instance attributes are
+                   interned.  we should try to do that here. */
+                if (NATIVE_CHECK_EXACT(d_key)) {
+                    Py_INCREF(d_key);
+                    INTERN_INPLACE(&d_key);
+                    Py_DECREF(d_key);
+                }
+                if (PyObject_SetItem(*dict, d_key, d_value) < 0)
+                    return NULL;
             }
-            if (PyObject_SetItem(*dict, d_key, d_value) < 0)
+        }
+        else
+        {
+            /* can happen that not a built-in dict is passed as state
+               fall back to iterating over items, instead of silently
+               failing with PyDict_Next */
+            items = PyMapping_Items(state);
+            if (items == NULL)
                 return NULL;
+            len = PySequence_Size(items);
+            if (len < 0)
+            {
+                Py_DECREF(items);
+                return NULL;
+            }
+            for ( i=0; i<len; ++i ) {
+                PyObject *item = PySequence_GetItem(items, i);
+                if (item == NULL)
+                {
+                    Py_DECREF(items);
+                    return NULL;
+                }
+                d_key = PyTuple_GetItem(item, 0);
+                if (d_key == NULL)
+                {
+                    Py_DECREF(item);
+                    Py_DECREF(items);
+                    return NULL;
+                }
+                d_value = PyTuple_GetItem(item, 1);
+                if (d_value == NULL)
+                {
+                    Py_DECREF(item);
+                    Py_DECREF(items);
+                    return NULL;
+                }
+
+                if (NATIVE_CHECK_EXACT(d_key)) {
+                    Py_INCREF(d_key);
+                    INTERN_INPLACE(&d_key);
+                    Py_DECREF(d_key);
+                }
+                Py_DECREF(item);
+                if (PyObject_SetItem(*dict, d_key, d_value) < 0)
+                {
+                    Py_DECREF(items);
+                    return NULL;
+                }
+            }
+            Py_DECREF(items);
         }
     }
 
@@ -666,6 +783,7 @@ Per__getstate__(cPersistentObject *self)
 static void
 Per_dealloc(cPersistentObject *self)
 {
+    PyObject_GC_UnTrack((PyObject *)self);
     if (self->state >= 0)
     {
         /* If the cache has been cleared, then a non-ghost object
